import os
import pytest
import shlex


@pytest.fixture(scope="session")
def run_wrapper():  # pragma: no cover
    v = os.environ.get("TINYSHAR_TEST_RUN_WRAPPER")

    if v is None:
        return []
    else:
        return shlex.split(v, posix=False)



import argparse
import contextlib
import sys
import tinyshar


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="create a self-extracting shell archive",
    )

    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s " + tinyshar.__version__
    )
    parser.add_argument(
        "-o",
        metavar="<file>",
        help="write output to <file> instead of stdout"
    )
    parser.add_argument(
        "-a",
        metavar="<dir>",
        action='append',
        default=[],
        help="directory containing files to be extracted to root. Can be specified multiple times"
    )
    parser.add_argument(
        "-r",
        metavar="<dir>",
        action='append',
        default=[],
        help="directory containing files to be extracted to arena. Can be specified multiple times"
    )
    parser.add_argument(
        "-p",
        metavar="<cmd>",
        action='append',
        default=[],
        help="command to execute before file extraction. Can be specified multiple times"
    )
    parser.add_argument(
        "-c",
        metavar="<cmd>",
        action='append',
        default=[],
        help="command to execute after file extraction. Can be specified multiple times"
    )
    parser.add_argument(
        "-L",
        action='store_true',
        default=False,
        help="follow directory symlinks"
    )
    parser.add_argument(
        "--no-shellcheck",
        action='store_true',
        help="do not check resulting archive with shellcheck"
    )
    parser.add_argument(
        "--no-tee",
        action='store_true',
        help="do tee stdout and stderr of extractor to file"
    )

    args = parser.parse_args(args=argv)

    shar = tinyshar.SharCreator()

    for i in args.a:
        shar.add_dir(i, '/', follow_symlinks=args.L)

    for i in args.r:
        shar.add_dir(i, '', follow_symlinks=args.L)

    for i in args.p:
        shar.add_pre(i)

    for i in args.c:
        shar.add_post(i)

    with contextlib.ExitStack() as exit_stack:
        if args.o:
            out_stm = exit_stack.enter_context(open(args.o, 'wb'))
        else:
            out_stm = sys.stdout.buffer

        try:
            shar.render(
                out_stm=out_stm,
                build_validators=[] if args.no_shellcheck else [tinyshar.ShellcheckValidator()],
                tee_to_file=not args.no_tee
            )
        except tinyshar.ValidatorError as e:
            sys.exit(str(e))



from invoke import task


@task
def clean(ctx):
    ctx.run("rm -rf dist")
    ctx.run("rm -rf .tox")
    ctx.run("rm -rf reports")
    ctx.run("rm -rf docs/build")


@task
def bump(ctx, part='patch'):
    ctx.run("bumpversion " + part)


@task
def check(ctx):
    ctx.run("flake8 --max-line-length=120 setup.py tasks.py src tests")


@task
def build(ctx):
    ctx.run("python setup.py sdist bdist_wheel")


@task
def test(ctx):
    ctx.run("tox")


@task
def docs(ctx):
    ctx.run("sphinx-build docs/source docs/build")



import binascii as _binascii
import collections as _collections
import contextlib as _contextlib
import hashlib as _hashlib
import io as _io
import os as _os
import posixpath as _posixpath
import shlex as _shlex
import shutil as _shutil
import subprocess as _subprocess
import sys as _sys


try:
    __version__ = __import__('pkg_resources').get_distribution(__name__).version
except:  # noqa # pragma: no cover
    pass


def _to_bytes(s):
    assert isinstance(s, (str, bytes))

    if isinstance(s, str):
        return s.encode()
    else:
        return s


def _checked_which(what):
    result = _shutil.which(what)
    if result is None:
        raise FileNotFoundError(what)
    return result


def _make_reader(what):
    if callable(what):
        what = what()

    if isinstance(what, str):
        what = what.encode()

    if isinstance(what, bytes):
        ofs = 0

        def reader(n):
            nonlocal ofs
            n = min(n, len(what) - ofs)
            result = what[ofs: ofs + n]
            ofs += n
            return result

        return reader

    if isinstance(what, _io.IOBase):
        def reader(n):
            result = what.read(n)

            if not result:
                what.close()

            return result

        return reader

    raise TypeError(type(what))


class Base64Encoder:
    _MAXBINSIZE = 57

    def encode(self, reader, writer):
        writer(b"base64 -d << _END_\n")

        while True:
            chunk = reader(self._MAXBINSIZE)
            if not chunk:
                break

            writer(_binascii.b2a_base64(chunk))

        writer(b"_END_\n")


class ValidatorError(RuntimeError):
    pass


@_contextlib.contextmanager
def ShellcheckValidator():
    process = _subprocess.Popen(
        [_checked_which('shellcheck'), '-'],
        stdin=_subprocess.PIPE,
        stdout=_sys.stderr
    )

    yield process.stdin.write

    process.stdin.close()
    rc = process.wait()
    if rc != 0:
        raise ValidatorError("shellcheck failed")


class Md5Validator:
    def __init__(self):
        self.hashes = []

    def wrap_reader(self, reader, fname):
        md5 = _hashlib.md5()

        def reader_wrapper(n):
            chunk = reader(n)

            if chunk:
                md5.update(chunk)
            else:
                self.hashes.append((fname, md5.hexdigest().encode()))

            return chunk

        return reader_wrapper

    def render(self, writer):
        writer(b"md5sum -c << _END_\n")
        for fname, md5 in self.hashes:
            writer(b"%s  %s\n" % (md5, fname))
        writer(b"_END_\n")


class SharCreator:
    def __init__(self):
        self._files = {}
        self._dirs = set()
        self._pre_chunks = []
        self._post_chunks = []
        self._files_by_tag = _collections.defaultdict(lambda: [])

    def files_by_tag(self, tag):
        return sorted(self._files_by_tag[tag])

    def files_by_tag_as_shell_str(self, tag):
        return ' '.join(_shlex.quote(i) for i in self.files_by_tag(tag))

    def add_file(self, name, content, tags=[]):
        assert isinstance(name, str)
        name = _posixpath.normpath(name)

        if name in ['.', '/']:
            raise IsADirectoryError(name)

        if name in self._files:
            raise FileExistsError(name)

        components = tuple(name.split('/'))
        is_abs = components[0] == ''
        for depth in range(1 + int(is_abs), len(components)):
            self._dirs.add(components[:depth])

        self._files[name] = content

        for tag in tags:
            assert isinstance(tag, str)
            self._files_by_tag[tag].append(name)

        return self

    def add_dir(self, src, dest, follow_symlinks=False):
        with _os.scandir(src) as it:
            for entry in it:
                src_name = _os.path.join(src, entry.name)
                dest_name = _posixpath.join(dest, entry.name)
                if entry.is_file():
                    self.add_file(dest_name, lambda fname=src_name: open(fname, 'rb'))
                elif entry.is_dir(follow_symlinks=follow_symlinks):
                    self.add_dir(src_name, dest_name)
                else:
                    raise ValueError("do not know how to deal with " + src_name)

        return self

    def _add_chunk(self, dest, chunk, order):
        assert isinstance(order, int)
        dest.append((order, chunk))
        return self

    def add_pre(self, chunk, *, order=0):
        return self._add_chunk(self._pre_chunks, chunk, order)

    def add_post(self, chunk, *, order=0):
        return self._add_chunk(self._post_chunks, chunk, order)

    def render(
        self,
        *,
        shebang=b'/bin/sh',
        out_stm=None,
        encoder=None,
        build_validators=None,
        extraction_validators=None,
        tee_to_file=True,
        _test_tmp_dir=None,
    ):
        """ Produce a shell script

        Note: it is ok to invoke this multiple times.

        Args:
            out_stm (Optional[IO[bytes]]):
            encoder (Optional[Encoder]):
            build_validators (Optional[List[BuildValidator]]):
            extraction_validators (Optional[List[ExtractionValidator]]):
            tee_to_file (Optional[bool]): Defaults to True.
            _test_tmp_dir: used by unittests to avoid leftoever temp directories

        Returns:
            List[bytes]: list of chunks comprising rendered result. Use "b''.join(render_result)" to obtain ``bytes``.
            None: if `out_stm` was supplied
        """
        encoder = encoder or Base64Encoder()
        if build_validators is None:
            build_validators = [ShellcheckValidator()]
        if extraction_validators is None:
            extraction_validators = [Md5Validator()]

        with _contextlib.ExitStack() as exit_stack:
            writers = []

            for validator in build_validators:
                writers.append(exit_stack.enter_context(validator))

            if out_stm is None:
                result_chunks = []
                writers.append(lambda what: result_chunks.append(what))
            else:
                writers.append(out_stm.write)

            def put(what):
                for writer in writers:
                    writer(what)

            def putnl():
                put(b"\n")

            def putl(what):
                put(what)
                putnl()

            def put_break():
                put(b"################################################################################\n")

            def put_annotation(s):
                putnl()
                put_break()
                put(b"# %s\n" % s)

            def put_chunks(annotation, chunks):
                if chunks:
                    put_annotation(annotation)

                    # according to https://wiki.python.org/moin/HowTo/Sorting/#Sort_Stability_and_Complex_Sorts
                    # sorts are guaranteed to stable.
                    for _, i in sorted(chunks, key=lambda i: i[0]):
                        putl(_to_bytes(i))

            put(b'#!%s\n' % _to_bytes(shebang))
            put(
                b'set -euxo pipefail\n'
                b'DIR=$('
            )

            if _test_tmp_dir is not None:
                put(b"TMPDIR=%s " % _to_bytes(_shlex.quote(_test_tmp_dir)))

            put(
                b'mktemp -d)\n'
                b'cd "$DIR"\n'
            )

            if tee_to_file:
                putl(b'{')

            put_chunks(b'PRE:', self._pre_chunks)

            files_map = []
            for i, (name, content) in enumerate(sorted(self._files.items())):
                tmp_name = b'%06d' % i
                files_map.append((tmp_name, name))

                put_annotation(b'file: %s\n' % name.encode())
                put(b'>"%s" ' % tmp_name)

                reader = _make_reader(content)

                for validator in extraction_validators:
                    reader = validator.wrap_reader(reader, tmp_name)

                encoder.encode(reader, put)

            if files_map:
                put_annotation(b"validation:\n")

                for validator in extraction_validators:
                    validator.render(put)

            put_break()
            put(
                b'mkdir arena\n'
                b'cd arena\n'
            )

            for i in sorted(self._dirs):
                put(b'mkdir -p %s\n' % _shlex.quote('/'.join(i)).encode())

            for tmp_name, name in files_map:
                target = _shlex.quote(name).encode()
                put(b"test '!' -d %s\n" % target)
                put(b"mv -f ../%s %s\n" % (tmp_name, target))

            put_chunks(b'POST:', self._post_chunks)

            if tee_to_file:
                putl(b'} 2>&1 | tee log')

            put_break()

            put(
                b"cd /\n"
                b'rm -rf "$DIR"\n'
            )

            put_break()

            if out_stm is None:
                return result_chunks



import pytest
from tinyshar._cli import main
import subprocess
import os
from contextlib import contextmanager


def test_empty():
    main([])


def test_validation_error():
    with pytest.raises(SystemExit):
        main(['-p', '"'])


def test_fail_due_symlink(tmpdir, monkeypatch):
    arena_dir = tmpdir / "arena"
    arena_dir.mkdir()

    # note: since there are no proper symlinks on Windows, we just
    # mock them on all platforms for testing purposes

    NAME = "6ac5f8bb-ac10-49af-b866-a1d21b42943c"
    (arena_dir / NAME).write_binary(b"")

    class MockEntry:  # pragma: no coverage
        def __init__(self, real):
            self._real = real
            self.name = real.name

        def is_file(self, follow_symlinks=True):
            if self.name == NAME:
                return False

            return self._real.is_file(follow_symlinks)

        def is_dir(self, follow_symlinks=True):
            if self.name == NAME:
                return False

            return self._real.is_file(follow_symlinks)

        def __enter__(self):
            return self._real.__enter__()

        def __leave__(self, *a):
            return self._real.__leave__(*a)

    real_scandir = os.scandir

    @contextmanager
    def mock_scandir(path):
        yield (MockEntry(i) for i in real_scandir(path))  # pragma: no coverage

    monkeypatch.setattr(os, "scandir", mock_scandir)

    with pytest.raises(ValueError) as e:
        main(["-r", str(arena_dir)])

    assert NAME in str(e.value)


def test_all(tmpdir, run_wrapper):
    root_dir = tmpdir / "root"
    # note: we assume that tmpdir is on C: drive on Windows
    root_root_dir = root_dir / os.path.splitdrive(tmpdir)[1] / "root_out"
    arena_dir = tmpdir / "arena"
    (root_root_dir / "file1").write_binary(b"text1", ensure=True)
    (root_root_dir / "dir" / "file1").write_binary(b"text2", ensure=True)
    (arena_dir / "file1").write_binary(b"text3", ensure=True)
    (arena_dir / "dir" / "file1").write_binary(b"text4", ensure=True)

    tmp_dir = tmpdir / "tmp"
    tmp_dir.mkdir()

    script_path = tmpdir / "script.sh"

    main([
        "-o", str(script_path),
        "-p", "true",
        "-p", "true",
        "-c", "true",
        "-c", "true",
        "-a", str(root_dir),
        "-r", str(arena_dir)
    ])

    script_path.chmod(0o500)

    env = dict(os.environ)
    env["TMPDIR"] = str(tmp_dir)
    subprocess.check_call(run_wrapper + [str(script_path)], env=env)



import os
import pytest
import tinyshar
import subprocess
import filecmp


@pytest.fixture
def shar():
    return tinyshar.SharCreator()


@pytest.fixture
def empty_path(tmpdir):
    saved_path = os.environ['PATH']
    empty = tmpdir / "empty"
    empty.mkdir()
    os.environ['PATH'] = str(empty)
    yield
    os.environ['PATH'] = saved_path


@pytest.fixture
def run(tmpdir, shar, run_wrapper):
    class Run:
        def __init__(self):
            self.run_count = 0
            self.pre_cb = []
            self.post_cb = []

        def add_pre(self, cb):
            self.pre_cb.append(cb)

        def add_post(self, cb):
            self.post_cb.append(cb)

        def __call__(
            self,
            *,
            expect_returncode=[0],
            cb_params={},
            tee_to_file=True,
            patch_cb=None
        ):
            # __tracebackhide__ = True
            assert self.run_count == 0
            self.run_count += 1

            for cb in self.pre_cb:
                cb(**cb_params)

            tmp_dir = tmpdir / "tmp"
            tmp_dir.mkdir()

            script_fname = str(tmpdir / "script.sh")

            with open(script_fname, "wb") as out_stm:
                render_opts = dict(
                    tee_to_file=tee_to_file,
                    _test_tmp_dir=str(tmp_dir)
                )
                if patch_cb:
                    rendered = b''.join(shar.render(**render_opts))
                    rendered = patch_cb(rendered)
                    out_stm.write(rendered)
                else:
                    shar.render(
                        out_stm=out_stm,
                        **render_opts
                    )

            os.chmod(script_fname, 0o500)

            cp = subprocess.run(
                run_wrapper + [script_fname],
                stdin=subprocess.DEVNULL,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )

            assert cp.returncode in expect_returncode
            tmp_list = tmp_dir.listdir()
            assert len(tmp_list) == (1 if cp.returncode else 0)

            if tmp_list:
                self.arena_dir = tmp_list[0] / 'arena'
            else:
                self.arena_dir = None

            for cb in self.post_cb:
                cb(returncode=cp.returncode, **cb_params)

    r = Run()
    yield r
    assert r.run_count == 1


@pytest.fixture
def cmpdirs(tmpdir, shar, run):
    dir_n = 0

    def add(dest):
        nonlocal dir_n
        d = tmpdir / ("%04d" % dir_n)
        d.mkdir()
        dir_n += 1

        def pre_run_cb(**kw):
            shar.add_dir(str(d), dest)

        def post_run_cb(expect_files=True, **kw):
            if os.path.isabs(dest):
                real_dest = dest
            else:
                if run.arena_dir is None:
                    return

                real_dest = os.path.join(run.arena_dir, dest)

            if expect_files:
                assert os.path.isdir(real_dest)
                diff = filecmp.dircmp(str(d), real_dest)
                assert len(diff.left_only) == 0
                assert len(diff.right_only) == 0
            else:
                assert not os.path.exists(run.arena_dir)

        run.add_pre(pre_run_cb)
        run.add_post(post_run_cb)

        return d

    return add


@pytest.fixture
def somefiles(tmpdir, cmpdirs):
    root_dir = cmpdirs(str(tmpdir / "root1"))
    (root_dir / "1").write_binary(b"one", ensure=True)
    (root_dir / "d1" / "11").write_binary(b"one-one", ensure=True)
    (root_dir / "d1" / "12").write_binary(b"one-two", ensure=True)
    arena_dir = cmpdirs("")
    (arena_dir / "2").write_binary(b"two", ensure=True)
    (arena_dir / "d2" / "2").write_binary(b"two-two", ensure=True)


def test_empty(shar):
    b''.join(shar.render())


def test_chaining(shar):
    (shar
        .add_pre("true")
        .add_post("true")
        .add_file("one", "")
        .render())


@pytest.mark.parametrize('extraction_validators', [[], None])
def test_files(shar, extraction_validators):
    shar.add_file("one", b"abc" * 10)
    shar.add_file("two/one", "")
    shar.add_file("/three", lambda: "")

    b''.join(shar.render(extraction_validators=extraction_validators))


def test_dup_file(shar):
    shar.add_file("one", '')
    with pytest.raises(FileExistsError):
        shar.add_file("one", '')


@pytest.mark.parametrize('name', ['/', '', '.'])
def test_is_a_directory_error1(shar, name):
    with pytest.raises(IsADirectoryError):
        shar.add_file(name, '')


def test_bad_contents_type(shar):
    shar.add_file("one", 1)

    with pytest.raises(TypeError):
        b''.join(shar.render())


def test_pre_post(shar):
    shar.add_pre('true')
    shar.add_pre(b'true')
    shar.add_post('false')
    shar.add_post(b'false')

    b''.join(shar.render())


def test_missing_shellcheck(shar, empty_path):
    with pytest.raises(FileNotFoundError):
        shar.render()


def test_shellcheck(shar):
    shar.add_pre('"')

    with pytest.raises(tinyshar.ValidatorError):
        b''.join(shar.render())


def test_no_shellcheck(shar):
    shar.add_pre('"')
    b''.join(shar.render(build_validators=[]))


@pytest.mark.parametrize('stage', ['add_pre', 'add_post'])
@pytest.mark.parametrize(
    'cmd,expect_returncode',
    [
        ('false', 1),
        ('true', 0),
        ('exit 42', 42),
    ]
)
@pytest.mark.parametrize('tee_to_file', [False, True])
def test_run_retcode(shar, run, stage, cmd, expect_returncode, tee_to_file):
    getattr(shar, stage)(cmd)
    run(expect_returncode=[expect_returncode], tee_to_file=tee_to_file)


@pytest.mark.parametrize('tee_to_file', [False, True])
def test_bad_md5(shar, run, tee_to_file):
    shar.add_file("one", "")
    run(
        expect_returncode=[1],
        patch_cb=lambda s: s.replace(b'd41d8cd98f00b204e9800998ecf8427e', b'deadbeaf'),
        tee_to_file=tee_to_file
    )


def test_dirs1(shar, run, somefiles):
    shar.add_post("false")
    run(expect_returncode=[1])


def test_dirs2(shar, run, somefiles):
    run()


def test_dirs3(shar, run, somefiles):
    shar.add_pre("false")
    run(expect_returncode=[1], cb_params=dict(expect_files=False))


def test_tags(shar):
    shar.add_file("one", "", tags=[])
    shar.add_file("two", "", tags=["x"])
    shar.add_file("\"three\"", "", tags=["x"])
    shar.add_file("four", "", tags=["y"])
    assert shar.files_by_tag_as_shell_str("z") == ""
    assert shar.files_by_tag_as_shell_str("x") == "'\"three\"' two"
    assert shar.files_by_tag_as_shell_str("y") == "four"


def test_chunk_order(shar):
    shar.add_pre("f", order=3)
    shar.add_pre("e", order=3)
    shar.add_pre(b"d")
    shar.add_pre("c")
    shar.add_pre("b", order=-1)
    shar.add_pre(b"a", order=-1)
    shar.add_post("xxx")

    s = b''.join(shar.render())
    assert b'\nb\na\nd\nc\nf\ne\n' in s


def test_chunk_order_assert(shar):
    with pytest.raises(AssertionError):
        shar.add_pre("f", order="x")


def test_target_is_a_dir(tmpdir, shar, run):
    d = tmpdir / "target"
    d.mkdir()
    shar.add_file(str(d), "")
    run(expect_returncode=[1])



from setuptools import setup, find_packages

setup(
    name="tinyshar",
    description="simple library and utility for creation of shell archives",
    long_description=open("README.md").read(),  # no "with..." will do for setup.py
    long_description_content_type='text/markdown; charset=UTF-8; variant=GFM',
    license="MIT",
    author="Kyrylo Shpytsya",
    author_email="kshpitsa@gmail.com",
    url="https://github.com/kshpytsya/tinyshar",
    setup_requires=["setuptools_scm"],
    use_scm_version=True,
    python_requires=">=3.6, <=3.7",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    entry_points={"console_scripts": ["tinyshar = tinyshar._cli:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: System :: Installation/Setup",
        "Topic :: System :: Software Distribution",
    ],
)
